/*
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2015 KYOCERA Corporation
 * (C) 2016 KYOCERA Corporation
 * (C) 2017 KYOCERA Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 and
 * only version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#define pr_fmt(fmt)	"BMS: %s: " fmt, __func__

#include <linux/module.h>
#include <linux/power_supply.h>
#include <oem-bms.h>

#define VBATT_SAMPLES			5
#define POWER_OFF_SOC			0
#define POWER_OFF_THRESHOLD_UV		(3300 * 1000)
#define VBATT_INITIAL_UV			(4200 * 1000)

static int vbatt_current_uv = VBATT_INITIAL_UV;
static int vbatt_buf_index = 0;
static int vbatt_buf[VBATT_SAMPLES] ={
	VBATT_INITIAL_UV, VBATT_INITIAL_UV, VBATT_INITIAL_UV, VBATT_INITIAL_UV, VBATT_INITIAL_UV
};

static int oem_bms_get_batt_prop(enum power_supply_property psp)
{
	union power_supply_propval ret = {0,};
	static struct power_supply *psy;

	psy = power_supply_get_by_name("battery");
	if (psy) {
		/* if battery has been registered, use the status property */
		psy->get_property(psy, psp, &ret);
		return ret.intval;
	}

	/* Default to false if the battery power supply is not registered. */
	pr_err("battery power supply is not registered\n");
	return -EINVAL;
}

static int oem_bms_get_average(int *buf, int num)
{
	int ave, cnt;
	int max = 0, min = 0, sum = 0;

	max = buf[0];
	min = buf[0];

	for (cnt = 0; cnt < num; cnt++) {
		if (max < buf[cnt]) {
			max = buf[cnt];
		} else if (min > buf[cnt]) {
			min = buf[cnt];
		}
		sum += buf[cnt];
	}

	if (num < 3) {
		ave = sum / num;
	} else {
		ave = (sum - max - min) / (num - 2);
	}

	pr_debug("ave:%d sum:%d max:%d min:%d num:%d\n",
		ave, sum, max, min, num);

	return ave;
}

static void oem_bms_update_vbatt_current(void)
{
	int vbatt_now_uv = oem_bms_get_batt_prop(POWER_SUPPLY_PROP_VOLTAGE_NOW);

	if (vbatt_now_uv == -EINVAL) {
		pr_err("invalid value vbatt_now_uv:%d\n", vbatt_now_uv);
		return;
	}
	vbatt_buf_index++;

	if (vbatt_buf_index == VBATT_SAMPLES) {
		vbatt_buf_index = 0;
	}
	vbatt_buf[vbatt_buf_index] = vbatt_now_uv;
	vbatt_current_uv = oem_bms_get_average(vbatt_buf, VBATT_SAMPLES);

	pr_debug("vbatt_buf [0]=%d [1]=%d [2]=%d [3]=%d [4]=%d vbatt_buf_index:%d\n",
		vbatt_buf[0], vbatt_buf[1], vbatt_buf[2], vbatt_buf[3], vbatt_buf[4], vbatt_buf_index);
}

static int oem_bms_check_low_vbatt(int soc, int battery_status)
{
	int result_soc = soc;

	if(vbatt_current_uv <= POWER_OFF_THRESHOLD_UV) {
		result_soc = POWER_OFF_SOC;
		pr_err("set auto power off soc:%d vbatt_uv:%d\n", soc, vbatt_current_uv);
	}
	pr_debug("result_soc:%d vbatt_current_uv:%d\n", result_soc, vbatt_current_uv);
	return result_soc;
}

int oem_bms_correct_soc(int in_soc)
{
	int result_soc = in_soc;
	int battery_status = oem_bms_get_batt_prop(POWER_SUPPLY_PROP_STATUS);

	if (battery_status == -EINVAL) {
		pr_err("invalid value battery_status:%d\n", battery_status);
		return result_soc;
	}

	oem_bms_update_vbatt_current();

	result_soc = oem_bms_check_low_vbatt(result_soc, battery_status);

	pr_debug("result_soc:%d in_soc:%d battery_status:%d\n", result_soc, in_soc, battery_status);

	return result_soc;
}
EXPORT_SYMBOL(oem_bms_correct_soc);

