/*
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2015 KYOCERA Corporation
 * (C) 2016 KYOCERA Corporation
 * (C) 2017 KYOCERA Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 and
 * only version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#define pr_fmt(fmt)	"OEM_CHG_DET %s: " fmt, __func__

#include <linux/module.h>
#include <linux/gpio.h>
#include <linux/power_supply.h>
#include <linux/interrupt.h>
#include <linux/of_gpio.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <linux/wakelock.h>

static void oem_chg_set_usb_det_n_irq(void);
static irqreturn_t oem_chg_usb_det_n_isr(int irq, void *ptr);

#define DRV_NAME "kc,oem-chg-det-driver"
#define DEV_NAME "USB_DET_N"
#define IRQ_NAME "USB_DET_N_IRQ"

#define USB_CHECK_COUNT   2
#define CHECK_CYCLE_MS   50

#define IS_USB_DET() (0 == gpio_get_value(oem_chg_det_dev->usb_det_n_gpio))

struct oem_chg_det_device {
	int usb_det_n_gpio;
	int usb_det_n_irq;
};

static struct oem_chg_det_device *oem_chg_det_dev;

static struct delayed_work usb_det_status_work;

static struct wake_lock usb_det_status_wake_lock;

static int usb_det_status = 0;

extern int msm_otg_change_charge_device(int);

static void usb_det_status_update(struct work_struct *work)
{
	static int check_cnt = 0;
	int new_usb_det_status = IS_USB_DET();

	if (usb_det_status == new_usb_det_status) {
		check_cnt = 0;
		pr_info("equ usb_status=%d\n", new_usb_det_status);
		goto set_irq;
	}
	check_cnt++;

	if (USB_CHECK_COUNT <= check_cnt) {
		pr_info("change usb_det_status old=%d new=%d\n", usb_det_status, new_usb_det_status);
		usb_det_status = new_usb_det_status;
		msm_otg_change_charge_device(usb_det_status);
		check_cnt = 0;
	}

	if (check_cnt) {
		schedule_delayed_work(&usb_det_status_work, msecs_to_jiffies(CHECK_CYCLE_MS));
		pr_info("set usb_det_status_work=%dms cnt=%d\n", CHECK_CYCLE_MS, check_cnt);
		return;
	}

set_irq:
	oem_chg_set_usb_det_n_irq();
	wake_unlock(&usb_det_status_wake_lock);
	pr_info("unlock usb_det_status_wake_lock\n");
	return;
}

static void oem_chg_set_usb_det_n_irq(void)
{
	int rc = 0;

	free_irq(oem_chg_det_dev->usb_det_n_irq, 0);
	if (usb_det_status) {
		rc = request_irq(oem_chg_det_dev->usb_det_n_irq, oem_chg_usb_det_n_isr, IRQF_TRIGGER_HIGH, IRQ_NAME, 0);
	} else {
		rc = request_irq(oem_chg_det_dev->usb_det_n_irq, oem_chg_usb_det_n_isr, IRQF_TRIGGER_LOW, IRQ_NAME, 0);
	}
	if (rc) {
		pr_err("couldn't register interrupts rc=%d\n", rc);
	}
}

static irqreturn_t oem_chg_usb_det_n_isr(int irq, void *ptr)
{
	disable_irq_nosync(oem_chg_det_dev->usb_det_n_irq);

	pr_info("USB_DET_N irq detected!! GPIO(%d) new_usb_det_status=%d\n", oem_chg_det_dev->usb_det_n_gpio, IS_USB_DET());

	wake_lock(&usb_det_status_wake_lock);
	schedule_delayed_work(&usb_det_status_work, msecs_to_jiffies(CHECK_CYCLE_MS));

	return IRQ_HANDLED;
}

static int oem_chg_det_probe(struct platform_device *pdev)
{
	int rc = 0;

	wake_lock_init(&usb_det_status_wake_lock, WAKE_LOCK_SUSPEND, "usb_detect");

	INIT_DELAYED_WORK(&usb_det_status_work, usb_det_status_update);

	if (!pdev->dev.of_node) {
		pr_err("No platform supplied from device tree.\n");
		rc = -EINVAL;
		goto err_arg;
	}

	oem_chg_det_dev = kzalloc(sizeof(struct oem_chg_det_device), GFP_KERNEL);
	if (!oem_chg_det_dev) {
		pr_err("kzalloc fail\n");
		rc = -ENOMEM;
		goto err_alloc;
	}

	oem_chg_det_dev->usb_det_n_gpio = of_get_named_gpio(pdev->dev.of_node, "usb-det-n-gpio", 0);
	if (oem_chg_det_dev->usb_det_n_gpio < 0) {
		pr_err("of_get_named_gpio failed.\n");
		rc = -EINVAL;
		goto err_gpio;
	}

	rc = gpio_request(oem_chg_det_dev->usb_det_n_gpio, DEV_NAME);
	if (rc) {
		pr_err("gpio_request failed.\n");
		goto err_gpio;
	}

	oem_chg_det_dev->usb_det_n_irq = gpio_to_irq(oem_chg_det_dev->usb_det_n_gpio);
	usb_det_status = IS_USB_DET();

	if (usb_det_status) {
		rc = request_irq(oem_chg_det_dev->usb_det_n_irq, oem_chg_usb_det_n_isr, IRQF_TRIGGER_HIGH, IRQ_NAME, 0);
	} else {
		rc = request_irq(oem_chg_det_dev->usb_det_n_irq, oem_chg_usb_det_n_isr, IRQF_TRIGGER_LOW, IRQ_NAME, 0);
	}
	if (rc) {
		pr_err("failed request_irq.\n");
		goto err_irq;
	}

	enable_irq_wake(oem_chg_det_dev->usb_det_n_irq);
	pr_info("initialization is completed!! GPIO(%d) usb_det_status=%d\n", oem_chg_det_dev->usb_det_n_gpio, usb_det_status);

	msm_otg_change_charge_device(usb_det_status);
	pr_info("Notify usb_det_status\n");
	return 0;

err_irq:
	gpio_free(oem_chg_det_dev->usb_det_n_gpio);

err_gpio:
	kfree(oem_chg_det_dev);

err_alloc:
err_arg:
	pr_err("failed.\n");

	return rc;
}

static int oem_chg_det_remove(struct platform_device *pdev)
{
	free_irq(oem_chg_det_dev->usb_det_n_irq, oem_chg_det_dev);
	gpio_free(oem_chg_det_dev->usb_det_n_gpio);
	kfree(oem_chg_det_dev);

	return 0;
}

static const struct of_device_id oem_chg_det_of_match[] = {
	{ .compatible = DRV_NAME, },
};

static int oem_chg_det_suspend(struct device *dev)
{
	return 0;
}

static int oem_chg_det_resume(struct device *dev)
{
	return 0;
}

static const struct dev_pm_ops oem_chg_det_pm_ops = {
	.suspend	= oem_chg_det_suspend,
	.resume		= oem_chg_det_resume,
};

static struct platform_driver oem_chg_det_driver = {
	.driver = {
		.name = DRV_NAME,
		.owner = THIS_MODULE,
		.of_match_table = oem_chg_det_of_match,
		.pm 	= &oem_chg_det_pm_ops,
	},
	.probe = oem_chg_det_probe,
	.remove = oem_chg_det_remove,
};

int __init oem_chg_det_init(void)
{
	return platform_driver_register(&oem_chg_det_driver);
}
module_init(oem_chg_det_init);

static void __exit oem_chg_det_exit(void)
{
	platform_driver_unregister(&oem_chg_det_driver);
}
module_exit(oem_chg_det_exit);

MODULE_AUTHOR("KYOCERA Corporation");
MODULE_DESCRIPTION("OEM CHARGER DETECT Driver");
MODULE_LICENSE("GPL");
MODULE_ALIAS("oem_chg_det");
